/* include tcp_connect */
#include	"unp.h"
/*
 * tcp_connect函数 根据host和serv建立tcp链接，函数内实现域名解析
 * 注：该函数没有做到容错性
 * 返回：若成功则为非负描述符，若失败则为-1
 */
int tcp_connect(const char *host, const char *serv) {
	int sockfd, n;
	struct addrinfo hints, *res, *ressave;

	bzero(&hints, sizeof(struct addrinfo));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	if ((n = getaddrinfo(host, serv, &hints, &res)) != 0)
		err_quit("tcp_connect error for %s, %s: %s", host, serv,
				gai_strerror(n));
	ressave = res;

	do {
		sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (sockfd < 0)
			continue; /* ignore this one */

		if (connect(sockfd, res->ai_addr, res->ai_addrlen) == 0)
			break; /* success */
		Close(sockfd); /* ignore this one */
	} while ((res = res->ai_next) != NULL);

	if (res == NULL) {/* errno set from final connect() */
		err_msg("tcp_connect error for %s, %s", host, serv);
		sockfd = -1;
	}

	freeaddrinfo(ressave);

	return (sockfd);
}
/* end tcp_connect */

/*
 * We place the wrapper function here, not in wraplib.c, because some
 * XTI programs need to include wraplib.c, and it also defines
 * a Tcp_connect() function.
 */

int Tcp_connect(const char *host, const char *serv) {
	return (tcp_connect(host, serv));
}
